#ifndef PRQUEUE_H
#define PRQUEUE_H

/** PRQueue - queue with position reservation for multi-threaded programs
** 
** Copyright (C) 2009  Eugene Surman
** You may distribute under the terms of the GNU General Public License
**
** class Mutex;
** class Lock;
** class DataQueue
** class PRQueue
** class SimpleQueue
** class Msg
** class StringMsg: public Msg
** class QueueTest
**/


using namespace std; // for clarity of the examples

#include <stdio.h>
#include <stdarg.h>
#include <sys/time.h>
#include <pthread.h>
#include <iostream>
#include <queue>


#define STR_MSG_SZ 512

//------------
class Mutex
{
public:
    friend class Lock;
    
    Mutex() {
        register int err;
        if( (err =pthread_mutex_init( &m_mux, NULL))) {
             cerr << "pthread_mutex_init: err=" << strerror(err) << " -exit" << endl;
             exit( 1);
        }
    }
    
    pthread_mutex_t* pmutex() { return &m_mux; }
    
protected:
    pthread_mutex_t m_mux;
    
    void lock() {
        register int err;
        if( (err =pthread_mutex_lock( &m_mux))) {
             cerr << "pthread_mutex_lock: err=" << strerror(err) << " -exit" << endl;
             exit( 1);
        }
    }
    
    void unlock() {
        register int err;
        if( (err =pthread_mutex_unlock( &m_mux))) {
             cerr << "pthread_mutex_unlock: err=" << strerror(err) << " -exit" << endl;
             exit( 1);
        }
    }
    
};//Mutex


//------------
class Lock
{
public:
    Lock( Mutex& m) : m_mux( m) { m_mux.lock(); }
    ~Lock(){ m_mux.unlock(); }
private:
    Mutex& m_mux;
};


template< typename DT> class PRQueue;

//------------
template< typename DT>
class DataQueue
{
public:
    friend class PRQueue<DT>;
    
    typedef typename deque< DT>::pointer data_pnt_t;
    typedef typename deque< bool>::pointer filled_pnt_t;
    
    struct position {
        position() : data_pnt(0), filled_pnt(0)	{}
	data_pnt_t data_pnt;
        filled_pnt_t filled_pnt;
    };
    
    bool pop( DT& data) {
        if( m_data_que.empty() || ! m_filled_que.front())
            return false;
        data = m_data_que.front();
        m_data_que.pop_front();
        m_filled_que.pop_front();
        return true;
    }
    
    void reserve( position& pos) {
        m_data_que.push_back( m_dummy);
        m_filled_que.push_back( false);
        pos.data_pnt = &m_data_que[ m_data_que.size() -1];
        pos.filled_pnt = &m_filled_que[ m_filled_que.size() -1];
    }
                
    void fill( const DT& data, const position& pos) {
        *pos.data_pnt = data;
        *pos.filled_pnt = true;
    }

    void push( const DT& data) {
        m_data_que.push_back( data);
        m_filled_que.push_back( true);
    }
    
    size_t size() const { return m_data_que.size(); } 
    bool empty() const { return m_data_que.empty(); } 
    
private :
    deque<DT> m_data_que;
    deque<bool> m_filled_que;
    DT m_dummy;
    
    DataQueue() {}
    
};//DataQueue


//------------
template < typename DT >
class PRQueue
{
public:
    typedef typename DataQueue<DT>::position position;

    PRQueue() { pthread_cond_init( &m_not_empty, NULL); }

    // Push data to queue using reserve position
    void push( const DT& data, const position& pos) {
        Lock lk( m_mux);
        m_que.fill( data, pos);
        notify_not_empty();
    } 
    
    // Pop data from input queue and reserve position in the output queue
    void pop( DT& data, PRQueue& outque, position& pos)
    {
        Lock lk( m_mux);
        while( true) {
            if( m_que.pop( data))
                break;
            wait_while_empty();
        }
        outque.reserve_pos( pos);
    }

    // Simple push
    void push( const DT& data) {
        Lock lk( m_mux);
        m_que.push( data);
        notify_not_empty();
    } 
    
    // Simple pop
    void pop( DT& data) {
        Lock lk( m_mux);
        while( true) {
            if( m_que.pop( data)) break;
            wait_while_empty();
        }
    }
    
    bool empty() {
        Lock lk( m_mux);
        return m_que.empty();
    }

    size_t size() {
        Lock lk( m_mux);
        return m_que.size();
    }
    
    Mutex& mux() { return m_mux; }
    
protected:
    DataQueue<DT> m_que;
        
    pthread_cond_t m_not_empty;
    Mutex m_mux;

    void reserve_pos( position& pos) {
        Lock lk( m_mux);
        m_que.reserve( pos);
    } 
    
    void wait_while_empty()  {
        pthread_cond_wait( &m_not_empty, m_mux.pmutex());
    }
    
    void notify_not_empty() {
        pthread_cond_signal( &m_not_empty);
    }
    
};//PRQueue


//------------
template< typename DT>
class SimpleQueue
{
public:
    typedef typename PRQueue<DT>::position position;

    SimpleQueue( size_t dummy =0) {
        pthread_cond_init( &m_not_empty, NULL);
    }
    
    void push( const DT& data, const position& pos) {
        Lock lk( m_mux);
        m_que.push( data);
        notify_not_empty();
    } 
    void push( const DT& data) {
        Lock lk( m_mux);
        m_que.push( data);
        notify_not_empty();
    } 
    
    void pop( DT& data, SimpleQueue& outque, position& pos) { // pos not used
        Lock lk( m_mux);
        wait_while_empty();
        data =m_que.front();
        m_que.pop();
    }
    void pop( DT& data) {
        Lock lk( m_mux);
        wait_while_empty();
        data =m_que.front();
        m_que.pop();
    }

    bool empty() {
        Lock lk( m_mux); return m_que.empty(); }
        
    size_t size() {
        Lock lk( m_mux); return m_que.size(); }
    
protected:
    queue< DT> m_que;
    pthread_cond_t m_not_empty;
    Mutex m_mux;

    void wait_while_empty() {
        while( m_que.empty()) pthread_cond_wait( &m_not_empty, m_mux.pmutex());
    }
        
    void notify_not_empty() { pthread_cond_signal( &m_not_empty);  }
    
};//SimpleQueue


//------------
class Msg
{
public:
    Msg() { gettimeofday( &m_time, 0); }
    Msg( const Msg& cp) : m_time( cp.m_time) {}
    virtual ~Msg() {}
    
    virtual void process( int arg) =0;
    virtual void print()  =0;
    
protected:
    timeval m_time;
    
};//Msg


//------------
class StringMsg: public Msg
{
public:
    StringMsg() {}
    
    StringMsg( const char* format, ...)  {
        va_list ap;
        va_start( ap, format);
        vsnprintf( m_buf, size(), format, ap);
    }

    StringMsg( const char* format, va_list ap)  {
        vsnprintf( m_buf, size(), format, ap);
    }
            
    void process( int arg);
    
    void print() {
        fputs( m_buf, stdout);  
        fflush( stdout);
    }

    size_t size() { return sizeof( m_buf); }
    
protected:    
    char m_buf[ STR_MSG_SZ+1];
    
}; //StringMsg


//------------
class QueueTest
{
public:
    friend class Msg;

    QueueTest( int qty_threads) : m_inited( false) {
        init( qty_threads);
    }

    size_t empty() { return input_que.empty() && output_que.empty(); }

#ifdef SIMPLE_QUE        
    SimpleQueue< Msg*> input_que, output_que; // try simple queue
#else
    PRQueue< Msg*> input_que, output_que;
#endif

    void push( const char* format, ...)  {
	va_list ap;
	va_start( ap, format);
	Msg* msg =new StringMsg( format, ap);
	input_que.push( msg);
	va_end( ap); 
    }
    	
protected:
    int m_qty_threads;
    bool m_inited;
            
    void init( int qty_threads);
    void start_input_thread();
    void start_output_thread();
    
};//QueueTest


#endif //PRQUEUE_H
